@@ -21,7 +21,11 @@
 #include <utility>
 #include <vector>
 
+#include "draco/material/material_library.h"
+#include "draco/scene/mesh_group.h"
+
 #ifdef DRACO_TRANSCODER_SUPPORTED
+#include "draco/core/constants.h"
 #include "draco/core/draco_test_base.h"
 #include "draco/core/draco_test_utils.h"
 #include "draco/core/draco_types.h"
@@ -226,8 +230,9 @@ TEST(GltfDecoderTest, AnimatedBonesGltf) {
 
   ASSERT_EQ(scene->NumMeshes(), 1);
   ASSERT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->NumMaterialIndices(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->GetMaterialIndex(0), 0);
+  const MeshGroup &mesh_group = *scene->GetMeshGroup(MeshGroupIndex(0));
+  ASSERT_EQ(mesh_group.NumMeshInstances(), 1);
+  ASSERT_EQ(mesh_group.GetMeshInstance(0).material_index, 0);
   ASSERT_EQ(scene->NumNodes(), 22);
   ASSERT_EQ(scene->NumRootNodes(), 1);
   ASSERT_EQ(scene->GetMaterialLibrary().NumMaterials(), 1);
@@ -253,8 +258,9 @@ TEST(GltfDecoderTest, AnimatedBonesGlb) {
   const std::unique_ptr<Scene> scene(DecodeGltfFileToScene(file_name));
   ASSERT_EQ(scene->NumMeshes(), 1);
   ASSERT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->NumMaterialIndices(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->GetMaterialIndex(0), 0);
+  const MeshGroup &mesh_group = *scene->GetMeshGroup(MeshGroupIndex(0));
+  ASSERT_EQ(mesh_group.NumMeshInstances(), 1);
+  ASSERT_EQ(mesh_group.GetMeshInstance(0).material_index, 0);
   ASSERT_EQ(scene->NumNodes(), 22);
   ASSERT_EQ(scene->NumRootNodes(), 1);
   ASSERT_EQ(scene->GetMaterialLibrary().NumMaterials(), 1);
@@ -384,8 +390,9 @@ TEST(GltfDecoderTest, SimpleScene) {
 
   ASSERT_EQ(scene->NumMeshes(), 1);
   ASSERT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->NumMaterialIndices(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->GetMaterialIndex(0), 0);
+  const MeshGroup &mesh_group = *scene->GetMeshGroup(MeshGroupIndex(0));
+  ASSERT_EQ(mesh_group.NumMeshInstances(), 1);
+  ASSERT_EQ(mesh_group.GetMeshInstance(0).material_index, 0);
   ASSERT_EQ(scene->NumNodes(), 2);
   ASSERT_EQ(scene->NumRootNodes(), 1);
   ASSERT_EQ(scene->GetMaterialLibrary().NumMaterials(), 1);
@@ -443,8 +450,9 @@ TEST(GltfDecoderTest, SimpleTriangleScene) {
 
   EXPECT_EQ(scene->NumMeshes(), 1);
   EXPECT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->NumMaterialIndices(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->GetMaterialIndex(0), 0);
+  const MeshGroup &mesh_group = *scene->GetMeshGroup(MeshGroupIndex(0));
+  ASSERT_EQ(mesh_group.NumMeshInstances(), 1);
+  ASSERT_EQ(mesh_group.GetMeshInstance(0).material_index, 0);
   EXPECT_EQ(scene->NumNodes(), 1);
   EXPECT_EQ(scene->NumRootNodes(), 1);
   EXPECT_EQ(scene->GetMaterialLibrary().NumMaterials(), 1);
@@ -744,8 +752,9 @@ TEST(GltfDecoderTest, GlbTextureSource) {
   const std::unique_ptr<Scene> scene(DecodeGltfFileToScene(file_name));
   EXPECT_EQ(scene->NumMeshes(), 1);
   EXPECT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->NumMaterialIndices(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->GetMaterialIndex(0), 0);
+  const MeshGroup &mesh_group = *scene->GetMeshGroup(MeshGroupIndex(0));
+  ASSERT_EQ(mesh_group.NumMeshInstances(), 1);
+  ASSERT_EQ(mesh_group.GetMeshInstance(0).material_index, 0);
   EXPECT_EQ(scene->NumNodes(), 3);
   EXPECT_EQ(scene->NumRootNodes(), 1);
   EXPECT_EQ(scene->GetMaterialLibrary().NumMaterials(), 1);
@@ -767,8 +776,9 @@ TEST(GltfDecoderTest, GltfTextureSource) {
   const std::unique_ptr<Scene> scene(DecodeGltfFileToScene(file_name));
   EXPECT_EQ(scene->NumMeshes(), 1);
   EXPECT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->NumMaterialIndices(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(MeshGroupIndex(0))->GetMaterialIndex(0), 0);
+  const MeshGroup &mesh_group = *scene->GetMeshGroup(MeshGroupIndex(0));
+  ASSERT_EQ(mesh_group.NumMeshInstances(), 1);
+  ASSERT_EQ(mesh_group.GetMeshInstance(0).material_index, 0);
   EXPECT_EQ(scene->NumNodes(), 3);
   EXPECT_EQ(scene->NumRootNodes(), 1);
   EXPECT_EQ(scene->GetMaterialLibrary().NumMaterials(), 1);
@@ -851,7 +861,7 @@ TEST(GltfDecoderTest, SimpleSkin) {
   // Check scene size.
   ASSERT_EQ(scene->NumMeshes(), 1);
   ASSERT_EQ(scene->NumMeshGroups(), 1);
-  ASSERT_EQ(scene->GetMeshGroup(draco::MeshGroupIndex(0))->NumMaterialIndices(),
+  ASSERT_EQ(scene->GetMeshGroup(draco::MeshGroupIndex(0))->NumMeshInstances(),
             1);
   ASSERT_EQ(scene->NumNodes(), 3);
   ASSERT_EQ(scene->NumRootNodes(), 1);
@@ -1156,7 +1166,7 @@ TEST(GltfDecoderTest, DecodeLightsIntoScene) {
   ASSERT_EQ(light.GetType(), draco::Light::POINT);
   ASSERT_EQ(light.GetRange(), 1000);
   ASSERT_EQ(light.GetInnerConeAngle(), 0.0);
-  ASSERT_NEAR(light.GetOuterConeAngle(), M_PI / 4.0f, 1e-8);
+  ASSERT_NEAR(light.GetOuterConeAngle(), DRACO_PI / 4.0f, 1e-8);
 
   // Check directional light with some properties specified.
   light = *scene->GetLight(LightIndex(2));
@@ -1174,7 +1184,7 @@ TEST(GltfDecoderTest, DecodeLightsIntoScene) {
   ASSERT_EQ(light.GetType(), draco::Light::SPOT);
   ASSERT_EQ(light.GetRange(), std::numeric_limits<float>::max());
   ASSERT_EQ(light.GetInnerConeAngle(), 0.0);
-  ASSERT_NEAR(light.GetOuterConeAngle(), M_PI / 4.0f, 1e-8);
+  ASSERT_NEAR(light.GetOuterConeAngle(), DRACO_PI / 4.0f, 1e-8);
 
   // Check that lights are referenced by the scene nodes.
   ASSERT_EQ(scene->GetNode(SceneNodeIndex(0))->GetLightIndex(),
@@ -1185,5 +1195,43 @@ TEST(GltfDecoderTest, DecodeLightsIntoScene) {
   ASSERT_EQ(scene->GetNode(SceneNodeIndex(4))->GetLightIndex(), LightIndex(1));
 }
 
+TEST(GltfDecoderTest, MaterialsVariants) {
+  // Checks that a model with KHR_materials_variants extension can be decoded.
+  draco::GltfDecoder decoder;
+  DRACO_ASSIGN_OR_ASSERT(auto scene,
+                         decoder.DecodeFromFileToScene(GetTestFileFullPath(
+                             "KhronosSampleModels/DragonAttenuation/glTF/"
+                             "DragonAttenuation.gltf")));
+  ASSERT_NE(scene, nullptr);
+  const draco::MaterialLibrary &library = scene->GetMaterialLibrary();
+  ASSERT_EQ(library.NumMaterialsVariants(), 2);
+  ASSERT_EQ(library.GetMaterialsVariantName(0), "Attenuation");
+  ASSERT_EQ(library.GetMaterialsVariantName(1), "Surface Color");
+
+  // Check that the cloth mesh has no material variants.
+  const draco::MeshGroup &cloth_group =
+      *scene->GetMeshGroup(draco::MeshGroupIndex(0));
+  ASSERT_EQ(cloth_group.GetName(), "Cloth Backdrop");
+  ASSERT_EQ(cloth_group.NumMeshInstances(), 1);
+  const auto &cloth_mappings =
+      cloth_group.GetMeshInstance(0).materials_variants_mappings;
+  ASSERT_EQ(cloth_mappings.size(), 0);
+
+  // Check that the dragon has correct materials variants.
+  const draco::MeshGroup &dragon_group =
+      *scene->GetMeshGroup(draco::MeshGroupIndex(1));
+  ASSERT_EQ(dragon_group.GetName(), "Dragon");
+  ASSERT_EQ(dragon_group.NumMeshInstances(), 1);
+  const auto &dragon_mappings =
+      dragon_group.GetMeshInstance(0).materials_variants_mappings;
+  ASSERT_EQ(dragon_mappings.size(), 2);
+  ASSERT_EQ(dragon_mappings[0].material, 1);
+  ASSERT_EQ(dragon_mappings[1].material, 2);
+  ASSERT_EQ(dragon_mappings[0].variants.size(), 1);
+  ASSERT_EQ(dragon_mappings[1].variants.size(), 1);
+  ASSERT_EQ(dragon_mappings[0].variants[0], 0);
+  ASSERT_EQ(dragon_mappings[1].variants[0], 1);
+}
+
 }  // namespace draco
 #endif  // DRACO_TRANSCODER_SUPPORTED
