@@ -15,7 +15,7 @@
 #include "draco/mesh/mesh_splitter.h"
 
 #ifdef DRACO_TRANSCODER_SUPPORTED
-#include "draco/mesh/triangle_soup_mesh_builder.h"
+#include <utility>
 
 namespace draco {
 
@@ -40,10 +40,11 @@ StatusOr<MeshSplitter::MeshVector> MeshSplitter::SplitMesh(
       split_attribute->attribute_type() == GeometryAttribute::MATERIAL;
 
   const int num_out_meshes = split_attribute->size();
+  WorkData work_data;
+  work_data.num_sub_mesh_faces.resize(num_out_meshes, 0);
 
   // Verify that the attribute values are defined "per-face", i.e., all points
   // on a face are always mapped to the same attribute value.
-  std::vector<int> num_sub_mesh_faces(num_out_meshes, 0);
   for (FaceIndex fi(0); fi < mesh.num_faces(); ++fi) {
     const auto face = mesh.face(fi);
     const AttributeValueIndex avi = split_attribute->mapped_index(face[0]);
@@ -53,36 +54,26 @@ StatusOr<MeshSplitter::MeshVector> MeshSplitter::SplitMesh(
                       "Attribute values not consistent on a face.");
       }
     }
-    num_sub_mesh_faces[avi.value()] += 1;
+    work_data.num_sub_mesh_faces[avi.value()] += 1;
   }
 
   // Create the sub-meshes.
-  std::vector<TriangleSoupMeshBuilder> mesh_builders(num_out_meshes);
+  work_data.mesh_builders.resize(num_out_meshes);
   // Map between attribute ids of the input and output meshes.
-  std::vector<int> att_id_map(mesh.num_attributes(), -1);
+  work_data.att_id_map.resize(mesh.num_attributes(), -1);
+  const int ignored_att_id =
+      (!preserve_split_attribute ? split_attribute_id : -1);
   for (int mi = 0; mi < num_out_meshes; ++mi) {
-    if (num_sub_mesh_faces[mi] == 0) {
+    if (work_data.num_sub_mesh_faces[mi] == 0) {
       continue;  // Empty mesh, don't initialize it.
     }
 
-    const int num_faces = num_sub_mesh_faces[mi];
-    mesh_builders[mi].Start(num_faces);
-    mesh_builders[mi].SetName(mesh.GetName());
-
-    // Add all attributes.
-    for (int ai = 0; ai < mesh.num_attributes(); ++ai) {
-      if (ai == split_attribute_id && !preserve_split_attribute) {
-        continue;
-      }
-      const GeometryAttribute *const src_att = mesh.attribute(ai);
-      att_id_map[ai] = mesh_builders[mi].AddAttribute(src_att->attribute_type(),
-                                                      src_att->num_components(),
-                                                      src_att->data_type());
-    }
+    const int num_faces = work_data.num_sub_mesh_faces[mi];
+    InitializeMeshBuilder(mi, num_faces, mesh, ignored_att_id, &work_data);
 
     // Reset the face counter for the sub-mesh. It will be used to keep track of
     // number of faces added to the sub-mesh.
-    num_sub_mesh_faces[mi] = 0;
+    work_data.num_sub_mesh_faces[mi] = 0;
   }
 
   // Go over all faces of the input mesh and add them to the appropriate
@@ -90,27 +81,87 @@ StatusOr<MeshSplitter::MeshVector> MeshSplitter::SplitMesh(
   for (FaceIndex fi(0); fi < mesh.num_faces(); ++fi) {
     const auto face = mesh.face(fi);
     const int sub_mesh_id = split_attribute->mapped_index(face[0]).value();
-    const FaceIndex target_face(num_sub_mesh_faces[sub_mesh_id]++);
-    for (int ai = 0; ai < mesh.num_attributes(); ++ai) {
-      if (ai == split_attribute_id && !preserve_split_attribute) {
-        continue;
-      }
-      const PointAttribute *const src_att = mesh.attribute(ai);
-      const int target_att_id = att_id_map[ai];
-      mesh_builders[sub_mesh_id].SetAttributeValuesForFace(
-          target_att_id, target_face, src_att->GetAddressOfMappedIndex(face[0]),
-          src_att->GetAddressOfMappedIndex(face[1]),
-          src_att->GetAddressOfMappedIndex(face[2]));
-    }
+    const FaceIndex target_fi(work_data.num_sub_mesh_faces[sub_mesh_id]++);
+    AddFaceToMeshBuilder(sub_mesh_id, fi, target_fi, mesh, &work_data);
   }
 
+  return FinalizeMeshes(mesh, &work_data);
+}
+
+StatusOr<MeshSplitter::MeshVector> MeshSplitter::SplitMeshToComponents(
+    const Mesh &mesh, const MeshConnectedComponents &connected_components) {
+  // Create the sub-meshes.
+  const int num_out_meshes = connected_components.NumConnectedComponents();
+  WorkData work_data;
+  work_data.mesh_builders.resize(num_out_meshes);
+  work_data.num_sub_mesh_faces.resize(num_out_meshes, 0);
+  work_data.att_id_map.resize(mesh.num_attributes(), -1);
+  for (int mi = 0; mi < num_out_meshes; ++mi) {
+    const int num_faces = connected_components.NumConnectedComponentFaces(mi);
+    work_data.num_sub_mesh_faces[mi] = num_faces;
+    InitializeMeshBuilder(mi, num_faces, mesh, -1, &work_data);
+  }
+
+  // Go over all faces of the input mesh and add them to the appropriate
+  // sub-mesh.
+  for (int mi = 0; mi < num_out_meshes; ++mi) {
+    for (int cfi = 0; cfi < connected_components.NumConnectedComponentFaces(mi);
+         ++cfi) {
+      const FaceIndex fi(
+          connected_components.GetConnectedComponent(mi).faces[cfi]);
+      const FaceIndex target_fi(cfi);
+      AddFaceToMeshBuilder(mi, fi, target_fi, mesh, &work_data);
+    }
+  }
+  return FinalizeMeshes(mesh, &work_data);
+}
+
+void MeshSplitter::InitializeMeshBuilder(int mb_index, int num_faces,
+                                         const Mesh &mesh,
+                                         int ignored_attribute_id,
+                                         WorkData *work_data) const {
+  work_data->mesh_builders[mb_index].Start(num_faces);
+  work_data->mesh_builders[mb_index].SetName(mesh.GetName());
+
+  // Add all attributes.
+  for (int ai = 0; ai < mesh.num_attributes(); ++ai) {
+    if (ai == ignored_attribute_id) {
+      continue;
+    }
+    const GeometryAttribute *const src_att = mesh.attribute(ai);
+    work_data->att_id_map[ai] = work_data->mesh_builders[mb_index].AddAttribute(
+        src_att->attribute_type(), src_att->num_components(),
+        src_att->data_type());
+  }
+}
+
+void MeshSplitter::AddFaceToMeshBuilder(int mb_index, FaceIndex source_fi,
+                                        FaceIndex target_fi, const Mesh &mesh,
+                                        WorkData *work_data) const {
+  const auto face = mesh.face(source_fi);
+  for (int ai = 0; ai < mesh.num_attributes(); ++ai) {
+    const PointAttribute *const src_att = mesh.attribute(ai);
+    const int target_att_id = work_data->att_id_map[ai];
+    if (target_att_id == -1) {
+      continue;
+    }
+    work_data->mesh_builders[mb_index].SetAttributeValuesForFace(
+        target_att_id, target_fi, src_att->GetAddressOfMappedIndex(face[0]),
+        src_att->GetAddressOfMappedIndex(face[1]),
+        src_att->GetAddressOfMappedIndex(face[2]));
+  }
+}
+
+StatusOr<MeshSplitter::MeshVector> MeshSplitter::FinalizeMeshes(
+    const Mesh &mesh, WorkData *work_data) const {
   // Finalize meshes.
+  const int num_out_meshes = work_data->mesh_builders.size();
   MeshVector out_meshes(num_out_meshes);
   for (int mi = 0; mi < num_out_meshes; ++mi) {
-    if (num_sub_mesh_faces[mi] == 0) {
-      continue;  // Empty mesh, don't create it.
+    if (work_data->num_sub_mesh_faces[mi] == 0) {
+      continue;
     }
-    out_meshes[mi] = mesh_builders[mi].Finalize();
+    out_meshes[mi] = work_data->mesh_builders[mi].Finalize();
     if (out_meshes[mi] == nullptr) {
       continue;
     }
@@ -128,7 +179,7 @@ StatusOr<MeshSplitter::MeshVector> MeshSplitter::SplitMesh(
 
     // Copy over attribute unique ids.
     for (int att_id = 0; att_id < mesh.num_attributes(); ++att_id) {
-      const int mapped_att_id = att_id_map[att_id];
+      const int mapped_att_id = work_data->att_id_map[att_id];
       if (mapped_att_id == -1) {
         continue;
       }
